from pubnub.endpoints.endpoint import Endpoint


class FileOperationEndpoint(Endpoint):
    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channel = None

    def channel(self, channel):
        self._channel = channel
        return self

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout



class PNMessageAction(object):
    def __init__(self, message_action=None):
        if message_action is not None:
            self.type = message_action['type']
            self.value = message_action['value']
            self.message_timetoken = message_action['messageTimetoken']
            self.uuid = message_action['uuid']
            self.action_timetoken = message_action['actionTimetoken']
        else:
            self.type = None
            self.value = None
            self.message_timetoken = None
            self.uuid = None
            self.action_timetoken = None

    def __str__(self):
        return "Message action with tt: %s for uuid %s with value %s " % (self.action_timetoken, self.uuid, self.value)


class PNGetMessageActionsResult(object):
    def __init__(self, result):
        """
        Representation of get message actions server response

        :param result: result of get message actions operation
        """
        self._result = result
        self.actions = result['actions']

    def __str__(self):
        return "Get message actions success"


class PNAddMessageActionResult(PNMessageAction):

    def __init__(self, message_action):
        super(PNAddMessageActionResult, self).__init__(message_action)


class PNRemoveMessageActionResult(object):
    def __init__(self, result):
        """s
        Representation of remove message actions server response

        :param result: result of remove message actions operation
        """
        self._result = result

    def __str__(self):
        return "Remove message actions success"




class PNPushAddChannelResult(object):
    def __str__(self):
        return "Channel successfully added"


class PNPushRemoveChannelResult(object):
    def __str__(self):
        return "Channel successfully removed"


class PNPushRemoveAllChannelsResult(object):
    def __str__(self):
        return "All channels successfully removed"


class PNPushListProvisionsResult(object):
    def __init__(self, channels):
        self.channels = channels

    def __str__(self):
        return "Push notification enabled on following channels: %s" % \
               ", ".join(self.channels)



from pubnub.endpoints.file_operations.file_based_endpoint import FileOperationEndpoint

from pubnub.crypto import PubNubFileCrypto
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.file import PNSendFileResult
from pubnub.endpoints.file_operations.publish_file_message import PublishFileMessage
from pubnub.endpoints.file_operations.fetch_upload_details import FetchFileUploadS3Data
from pubnub.request_handlers.requests_handler import RequestsRequestHandler
from pubnub.endpoints.mixins import TimeTokenOverrideMixin


class SendFileNative(FileOperationEndpoint, TimeTokenOverrideMixin):
    def __init__(self, pubnub):
        super(SendFileNative, self).__init__(pubnub)
        self._file_name = None
        self._pubnub = pubnub
        self._file_upload_envelope = None
        self._message = None
        self._should_store = None
        self._ttl = 0
        self._meta = None
        self._cipher_key = None
        self._file_object = None
        self._replicate = None
        self._ptto = None

    def file_object(self, fd):
        self._file_object = fd
        return self

    def build_params_callback(self):
        return lambda a: {}

    def build_path(self):
        return self._file_upload_envelope.result.data["url"]

    def encrypt_payload(self):
        if self._cipher_key or self._pubnub.config.cipher_key:
            try:
                payload = self._file_object.read()
            except AttributeError:
                payload = self._file_object

            return PubNubFileCrypto(self._pubnub.config).encrypt(
                self._cipher_key or self._pubnub.config.cipher_key,
                payload
            )
        else:
            return self._file_object

    def build_file_upload_request(self):
        file = self.encrypt_payload()
        multipart_body = {}
        for form_field in self._file_upload_envelope.result.data["form_fields"]:
            multipart_body[form_field["key"]] = (None, form_field["value"])

        multipart_body["file"] = (self._file_name, file, None)

        return multipart_body

    def http_method(self):
        return HttpMethod.POST

    def custom_params(self):
        return {}

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()
        self.validate_file_object()
        self.validate_file_name()

    def use_base_path(self):
        return False

    def non_json_response(self):
        return True

    def is_auth_required(self):
        return False

    def should_store(self, should_store):
        self._should_store = bool(should_store)
        return self

    def ttl(self, ttl):
        self._ttl = ttl
        return self

    def meta(self, meta):
        self._meta = meta
        return self

    def message(self, message):
        self._message = message
        return self

    def file_name(self, file_name):
        self._file_name = file_name
        return self

    def cipher_key(self, cipher_key):
        self._cipher_key = cipher_key
        return self

    def create_response(self, envelope, data=None):
        return PNSendFileResult(envelope, self._file_upload_envelope)

    def operation_type(self):
        return PNOperationType.PNSendFileAction

    def request_headers(self):
        return {}

    def name(self):
        return "Send file to S3"

    def sync(self):
        self._file_upload_envelope = FetchFileUploadS3Data(self._pubnub).\
            channel(self._channel).\
            file_name(self._file_name).sync()

        response_envelope = super(SendFileNative, self).sync()

        publish_file_response = PublishFileMessage(self._pubnub).\
            channel(self._channel).\
            meta(self._meta).\
            message(self._message).\
            file_id(response_envelope.result.file_id).\
            file_name(response_envelope.result.name).\
            should_store(self._should_store).\
            ttl(self._ttl).\
            replicate(self._replicate).\
            ptto(self._ptto).\
            cipher_key(self._cipher_key).sync()

        response_envelope.result.timestamp = publish_file_response.result.timestamp
        return response_envelope

    def pn_async(self, callback):
        return RequestsRequestHandler(self._pubnub).async_file_based_operation(self.sync, callback, "File Download")



from pubnub.models.consumer.v3.pn_resource import PNResource


class User(PNResource):

    def __init__(self, resource_name=None, resource_pattern=None):
        super(User, self).__init__(resource_name, resource_pattern)

    @staticmethod
    def id(user_id):
        user = User(resource_name=user_id)
        return user

    @staticmethod
    def pattern(user_pattern):
        user = User(resource_pattern=user_pattern)
        return user

    def read(self):
        self._read = True
        return self

    def write(self):
        self._write = True
        return self

    def create(self):
        self._create = True
        return self

    def manage(self):
        self._manage = True
        return self

    def delete(self):
        self._delete = True
        return self



class HttpMethod(object):
    GET = 1
    POST = 2
    DELETE = 3
    PATCH = 4

    @classmethod
    def string(cls, method):
        if method == cls.GET:
            return "GET"
        elif method == cls.POST:
            return "POST"
        elif method == cls.DELETE:
            return "DELETE"
        elif method == cls.PATCH:
            return "PATCH"


class PNStatusCategory(object):
    PNUnknownCategory = 1
    PNAcknowledgmentCategory = 2
    PNAccessDeniedCategory = 3
    PNTimeoutCategory = 4
    PNNetworkIssuesCategory = 5
    PNConnectedCategory = 6
    PNReconnectedCategory = 7
    PNDisconnectedCategory = 8
    PNUnexpectedDisconnectCategory = 9
    PNCancelledCategory = 10
    PNBadRequestCategory = 11
    PNMalformedFilterExpressionCategory = 12
    PNMalformedResponseCategory = 13
    PNDecryptionErrorCategory = 14
    PNTLSConnectionFailedCategory = 15
    PNTLSUntrustedCertificateCategory = 16
    PNInternalExceptionCategory = 17


class PNOperationType(object):
    PNSubscribeOperation = 1
    PNUnsubscribeOperation = 2
    PNPublishOperation = 3
    PNHistoryOperation = 4
    PNWhereNowOperation = 5

    PNHeartbeatOperation = 6
    PNSetStateOperation = 7
    PNAddChannelsToGroupOperation = 8
    PNRemoveChannelsFromGroupOperation = 9
    PNChannelGroupsOperation = 10
    PNRemoveGroupOperation = 11
    PNChannelsForGroupOperation = 12
    PNPushNotificationEnabledChannelsOperation = 13
    PNAddPushNotificationsOnChannelsOperation = 14
    PNRemovePushNotificationsFromChannelsOperation = 15
    PNRemoveAllPushNotificationsOperation = 16
    PNTimeOperation = 17

    PNHereNowOperation = 18
    PNGetState = 19
    PNAccessManagerAudit = 20
    PNAccessManagerGrant = 21
    PNAccessManagerRevoke = 22
    PNHistoryDeleteOperation = 23
    PNMessageCountOperation = 24
    PNFireOperation = 25
    PNSignalOperation = 26

    PNAccessManagerGrantToken = 41
    PNAddMessageAction = 42
    PNGetMessageActions = 43
    PNDeleteMessageAction = 44
    PNFetchMessagesOperation = 45

    PNGetFilesAction = 46
    PNDeleteFileOperation = 47
    PNGetFileDownloadURLAction = 48
    PNFetchFileUploadS3DataAction = 49
    PNDownloadFileAction = 50
    PNSendFileAction = 51
    PNSendFileNotification = 52

    PNSetUuidMetadataOperation = 53
    PNGetUuidMetadataOperation = 54
    PNRemoveUuidMetadataOperation = 55
    PNGetAllUuidMetadataOperation = 56

    PNSetChannelMetadataOperation = 57
    PNGetChannelMetadataOperation = 58
    PNRemoveChannelMetadataOperation = 59
    PNGetAllChannelMetadataOperation = 60

    PNSetChannelMembersOperation = 61
    PNGetChannelMembersOperation = 62
    PNRemoveChannelMembersOperation = 63
    PNManageChannelMembersOperation = 64

    PNSetMembershipsOperation = 65
    PNGetMembershipsOperation = 66
    PNRemoveMembershipsOperation = 67
    PNManageMembershipsOperation = 68


class PNHeartbeatNotificationOptions(object):
    NONE = 1
    FAILURES = 2
    ALL = 3


class PNReconnectionPolicy(object):
    NONE = 1
    LINEAR = 2
    EXPONENTIAL = 3


class PNPushType(object):
    APNS = 1
    MPNS = 2
    GCM = 3
    APNS2 = 4


class PNResourceType(object):
    CHANNEL = "channel"
    GROUP = "group"
    USER = "user"
    SPACE = "space"


class PNMatchType(object):
    RESOURCE = "resource"
    PATTERN = "pattern"


class PNPushEnvironment(object):
    DEVELOPMENT = "development"
    PRODUCTION = "production"



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.signal import PNSignalResult


class Signal(Endpoint):
    SIGNAL_PATH = '/signal/%s/%s/0/%s/0/%s'

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channel = None
        self._message = None

    def channel(self, channel):
        self._channel = str(channel)
        return self

    def message(self, message):
        self._message = message
        return self

    def build_path(self):
        stringified_message = utils.write_value_as_string(self._message)
        msg = utils.url_encode(stringified_message)
        return Signal.SIGNAL_PATH % (
            self.pubnub.config.publish_key, self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel), msg
        )

    def custom_params(self):
        return {}

    def http_method(self):
        return HttpMethod.GET

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_publish_key()
        self.validate_channel()

    def create_response(self, result):  # pylint: disable=W0221
        return PNSignalResult(result)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNSignalOperation

    def name(self):
        return "Signal"



from .enums import PNHeartbeatNotificationOptions, PNReconnectionPolicy
from . import utils


class PNConfiguration(object):
    DEFAULT_PRESENCE_TIMEOUT = 300
    DEFAULT_HEARTBEAT_INTERVAL = 280

    def __init__(self):
        # TODO: add validation
        self.uuid = None
        self.origin = "ps.pndsn.com"
        self.ssl = True
        self.non_subscribe_request_timeout = 10
        self.subscribe_request_timeout = 310
        self.connect_timeout = 10
        self.subscribe_key = None
        self.publish_key = None
        self.secret_key = None
        self.cipher_key = None
        self.auth_key = None
        self.filter_expression = None
        self.enable_subscribe = True
        self.crypto_instance = None
        self.file_crypto_instance = None
        self.log_verbosity = False
        self.enable_presence_heartbeat = False
        self.heartbeat_notification_options = PNHeartbeatNotificationOptions.FAILURES
        self.reconnect_policy = PNReconnectionPolicy.NONE
        self.daemon = False
        self.disable_token_manager = False
        self.use_random_initialization_vector = True
        self.suppress_leave_events = False

        self.heartbeat_default_values = True
        self._presence_timeout = PNConfiguration.DEFAULT_PRESENCE_TIMEOUT
        self._heartbeat_interval = PNConfiguration.DEFAULT_HEARTBEAT_INTERVAL

    def validate(self):
        assert self.uuid is None or isinstance(self.uuid, str)

        if not self.uuid:
            self.uuid = utils.uuid()

    def scheme(self):
        if self.ssl:
            return "https"
        else:
            return "http"

    def scheme_extended(self):
        return self.scheme() + "://"

    def scheme_and_host(self):
        return self.scheme_extended() + self.origin

    def set_presence_timeout_with_custom_interval(self, timeout, interval):
        self.heartbeat_default_values = False
        self._presence_timeout = timeout
        self._heartbeat_interval = interval

    def set_presence_timeout(self, timeout):
        self.set_presence_timeout_with_custom_interval(timeout, (timeout / 2) - 1)

    @property
    def crypto(self):
        if self.crypto_instance is None:
            self._init_cryptodome()

        return self.crypto_instance

    def _init_cryptodome(self):
        from .crypto import PubNubCryptodome
        self.crypto_instance = PubNubCryptodome(self)

    def _init_file_crypto(self):
        from .crypto import PubNubFileCrypto
        self.file_crypto_instance = PubNubFileCrypto(self)

    @property
    def file_crypto(self):
        if not self.file_crypto_instance:
            self._init_file_crypto()

        return self.file_crypto_instance

    @property
    def port(self):
        return 443 if self.ssl == "https" else 80

    @property
    def presence_timeout(self):
        return self._presence_timeout

    @property
    def heartbeat_interval(self):
        return self._heartbeat_interval

        # TODO: set log level
        # TODO: set log level



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.exceptions import PubNubException
from pubnub.errors import PNERR_MESSAGE_MISSING
from pubnub.models.consumer.pubsub import PNFireResult


class Fire(Endpoint):
    # /publish/<pub_key>/<sub_key>/<signature>/<channel>/<callback>/<message>[?argument(s)]
    FIRE_GET_PATH = "/publish/%s/%s/0/%s/%s/%s"
    FIRE_POST_PATH = "/publish/%s/%s/0/%s/%s"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channel = None
        self._message = None
        self._use_post = None
        self._meta = None

    def channel(self, channel):
        self._channel = str(channel)
        return self

    def message(self, message):
        self._message = message
        return self

    def use_post(self, use_post):
        self._use_post = bool(use_post)
        return self

    def meta(self, meta):
        self._meta = meta
        return self

    def build_data(self):
        if self._use_post is True:
            cipher = self.pubnub.config.cipher_key
            if cipher is not None:
                return '"' + self.pubnub.config.crypto.encrypt(cipher, utils.write_value_as_string(self._message)) + '"'
            else:
                return utils.write_value_as_string(self._message)
        else:
            return None

    def custom_params(self):
        params = {}
        if self._meta is not None:
            params['meta'] = utils.write_value_as_string(self._meta)
        params["store"] = "0"
        params["norep"] = "1"
        if self.pubnub.config.auth_key is not None:
            params["auth"] = utils.url_encode(self.pubnub.config.auth_key)
        return params

    def build_path(self):
        if self._use_post:
            return Fire.FIRE_POST_PATH % (self.pubnub.config.publish_key,
                                          self.pubnub.config.subscribe_key,
                                          utils.url_encode(self._channel), 0)
        else:
            cipher = self.pubnub.config.cipher_key
            stringified_message = utils.write_value_as_string(self._message)

            if cipher is not None:
                stringified_message = '"' + self.pubnub.config.crypto.encrypt(cipher, stringified_message) + '"'

            stringified_message = utils.url_encode(stringified_message)

            return Fire.FIRE_GET_PATH % (self.pubnub.config.publish_key,
                                         self.pubnub.config.subscribe_key,
                                         utils.url_encode(self._channel), 0, stringified_message)

    def http_method(self):
        if self._use_post is True:
            return HttpMethod.POST
        else:
            return HttpMethod.GET

    def validate_params(self):
        self.validate_channel()

        if self._message is None:
            raise PubNubException(pn_error=PNERR_MESSAGE_MISSING)

        self.validate_subscribe_key()
        self.validate_publish_key()

    def create_response(self, envelope):
        """
        :param envelope: an already serialized json response
        :return:
        """
        if envelope is None:
            return None

        timetoken = int(envelope[2])

        res = PNFireResult(envelope, timetoken)

        return res

    def is_auth_required(self):
        return True

    def affected_channels(self):
        return None

    def affected_channels_groups(self):
        return None

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNFireOperation

    def name(self):
        return "Fire"



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_PAM_NO_FLAGS, PNERR_PAM_INVALID_ARGUMENTS
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.access_manager import PNAccessManagerGrantResult


class Grant(Endpoint):
    GRANT_PATH = "/v2/auth/grant/sub-key/%s"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._auth_keys = []
        self._channels = []
        self._groups = []
        self._uuids = []
        self._read = None
        self._write = None
        self._manage = None
        self._delete = None
        self._ttl = None
        self._get = None
        self._update = None
        self._join = None

        self._sort_params = True

    def get(self, flag):
        self._get = flag
        return self

    def update(self, flag):
        self._update = flag
        return self

    def join(self, flag):
        self._join = flag
        return self

    def uuids(self, uuids):
        utils.extend_list(self._uuids, uuids)
        return self

    def auth_keys(self, auth_keys):
        utils.extend_list(self._auth_keys, auth_keys)
        return self

    def channels(self, channels):
        utils.extend_list(self._channels, channels)
        return self

    def channel_groups(self, channel_groups):
        utils.extend_list(self._groups, channel_groups)
        return self

    def read(self, flag):
        self._read = flag
        return self

    def write(self, flag):
        self._write = flag
        return self

    def manage(self, flag):
        self._manage = flag
        return self

    def delete(self, flag):
        self._delete = flag
        return self

    def ttl(self, ttl):
        self._ttl = ttl
        return self

    def encoded_params(self):
        params = {}

        if self._auth_keys:
            params['auth'] = utils.join_items_and_encode(self._auth_keys)

        if self._channels:
            params['channel'] = utils.join_channels(self._channels)

        if self._groups:
            params['channel-group'] = utils.join_items_and_encode(self._groups)

        return params

    def custom_params(self):
        params = {}

        if self._read is not None:
            params['r'] = '1' if self._read is True else '0'
        if self._write is not None:
            params['w'] = '1' if self._write is True else '0'
        if self._manage is not None:
            params['m'] = '1' if self._manage is True else '0'
        if self._delete is not None:
            params['d'] = '1' if self._delete is True else '0'
        if self._get is not None:
            params['g'] = '1' if self._get is True else '0'
        if self._update is not None:
            params['u'] = '1' if self._update is True else '0'
        if self._join is not None:
            params['j'] = '1' if self._join is True else '0'

        if self._auth_keys:
            params['auth'] = utils.join_items(self._auth_keys)

        if self._channels:
            params['channel'] = utils.join_items(self._channels)

        if self._groups:
            params['channel-group'] = utils.join_items(self._groups)

        if self._uuids:
            params['target-uuid'] = utils.join_items(self._uuids)

        if self._ttl is not None:
            params['ttl'] = str(int(self._ttl))

        return params

    def build_path(self):
        return Grant.GRANT_PATH % self.pubnub.config.subscribe_key

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_secret_key()
        self.validate_publish_key()
        # self.validate_channels_and_groups()

        if self._channels and self._groups and self._uuids:
            raise PubNubException(
                pn_error=PNERR_PAM_INVALID_ARGUMENTS,
                errormsg="Grants for channels or channelGroups can't be changed together with grants for UUIDs")

        if self._uuids and not self._auth_keys:
            raise PubNubException(pn_error=PNERR_PAM_INVALID_ARGUMENTS, errormsg="UUIDs grant management require "
                                                                                 "providing non empty authKeys"
                                  )

        if self._write is None and self._read is None and self._manage is None and self._get is None \
                and self._update is None and self._join is None:
            raise PubNubException(pn_error=PNERR_PAM_NO_FLAGS)

    def create_response(self, envelope):
        return PNAccessManagerGrantResult.from_json(envelope['payload'])

    def is_auth_required(self):
        return False

    def affected_channels(self):
        return self._channels

    def affected_channels_groups(self):
        return self._groups

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNAccessManagerGrant

    def name(self):
        return "Grant"



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.presence import PNGetStateResult
from pubnub.endpoints.mixins import UUIDValidatorMixin


class GetState(Endpoint, UUIDValidatorMixin):
    # /v2/presence/sub-key/<subscribe_key>/channel/<channel>/uuid/<uuid>/data?state=<state>
    GET_STATE_PATH = "/v2/presence/sub-key/%s/channel/%s/uuid/%s"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channels = []
        self._groups = []
        self._uuid = self.pubnub.uuid

    def channels(self, channels):
        utils.extend_list(self._channels, channels)
        return self

    def uuid(self, uuid):
        self._uuid = uuid
        return self

    def channel_groups(self, channel_groups):
        utils.extend_list(self._groups, channel_groups)
        return self

    def custom_params(self):
        params = {}

        if len(self._groups) > 0:
            params['channel-group'] = utils.join_items(self._groups)

        return params

    def build_path(self):
        return GetState.GET_STATE_PATH % (
            self.pubnub.config.subscribe_key,
            utils.join_channels(self._channels),
            utils.url_encode(self._uuid)
        )

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channels_and_groups()
        self.validate_uuid()

    def create_response(self, envelope):
        if len(self._channels) == 1 and len(self._groups) == 0:
            channels = {self._channels[0]: envelope['payload']}
        else:
            channels = envelope['payload']['channels']

        return PNGetStateResult(channels)

    def is_auth_required(self):
        return True

    def affected_channels(self):
        return self._channels

    def affected_channels_groups(self):
        return self._groups

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNGetState

    def name(self):
        return "GetState"



from abc import ABCMeta, abstractmethod
from .dtos import SubscribeOperation, UnsubscribeOperation
from . import utils


class PubSubBuilder(object):
    __metaclass__ = ABCMeta

    def __init__(self, subscription_manager):
        self._subscription_manager = subscription_manager
        self._channel_subscriptions = []
        self._channel_group_subscriptions = []

    # TODO: make the 'channel' alias
    def channels(self, channels_list):
        utils.extend_list(self._channel_subscriptions, channels_list)

        return self

    def channel_groups(self, channel_groups_list):
        utils.extend_list(self._channel_group_subscriptions, channel_groups_list)

        return self

    @abstractmethod
    def execute(self):
        pass


class SubscribeBuilder(PubSubBuilder):
    def __init__(self, subscription_manager):
        super(SubscribeBuilder, self).__init__(subscription_manager)
        self._presence_enabled = False
        self._timetoken = 0

    def with_presence(self):
        self._presence_enabled = True
        return self

    def with_timetoken(self, timetoken):
        self._timetoken = timetoken
        return self

    def channel_subscriptions(self):
        return self._channel_subscriptions

    def channel_group_subscriptions(self):
        return self._channel_group_subscriptions

    def execute(self):
        subscribe_operation = SubscribeOperation(
            channels=self._channel_subscriptions,
            channel_groups=self._channel_group_subscriptions,
            timetoken=self._timetoken,
            presence_enabled=self._presence_enabled
        )

        self._subscription_manager.adapt_subscribe_builder(subscribe_operation)


class UnsubscribeBuilder(PubSubBuilder):
    def execute(self):
        unsubscribe_operation = UnsubscribeOperation(
            channels=self._channel_subscriptions,
            channel_groups=self._channel_group_subscriptions
        )

        self._subscription_manager.adapt_unsubscribe_builder(unsubscribe_operation)



from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.time import PNTimeResponse


class Time(Endpoint):
    TIME_PATH = "/time/0"

    def custom_params(self):
        return {}

    def build_path(self):
        return Time.TIME_PATH

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        pass

    def is_auth_required(self):
        return False

    def create_response(self, envelope):
        return PNTimeResponse(envelope)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNTimeOperation

    def name(self):
        return "Time"



from enum import Enum


class PNSortKeyValue(Enum):
    ID = 1
    NAME = 2
    UPDATED = 3


class PNSortDirection(Enum):
    ASC = 1
    DESC = 2


class PNSortKey:
    def __init__(self, sort_key_value, direction):
        self._sort_key_value = sort_key_value
        self._direction = direction

    @staticmethod
    def asc(sort_key_value):
        return PNSortKey(sort_key_value, PNSortDirection.ASC)

    @staticmethod
    def desc(sort_key_value):
        return PNSortKey(sort_key_value, PNSortDirection.DESC)

    def key_str(self):
        if self._sort_key_value == PNSortKeyValue.ID:
            return "id"
        elif self._sort_key_value == PNSortKeyValue.NAME:
            return "name"
        elif self._sort_key_value == PNSortKeyValue.UPDATED:
            return "updated"
        else:
            raise ValueError()

    def dir_str(self):
        if self._direction == PNSortDirection.ASC:
            return "asc"
        elif self._direction == PNSortDirection.DESC:
            return "desc"
        else:
            raise ValueError()



from pubnub.endpoints.access.grant import Grant
from pubnub.enums import PNOperationType


class Revoke(Grant):
    def __init__(self, pubnub):
        Grant.__init__(self, pubnub)
        self._read = False
        self._write = False
        self._manage = False
        self._get = False
        self._update = False
        self._join = False

        self._sort_params = True

    def read(self, flag):
        raise NotImplementedError

    def write(self, flag):
        raise NotImplementedError

    def manage(self, flag):
        raise NotImplementedError

    def operation_type(self):
        return PNOperationType.PNAccessManagerRevoke

    def name(self):
        return "Revoke"



import datetime
import hmac
import json
import uuid as u
import threading
import urllib
from hashlib import sha256

from .enums import PNStatusCategory, PNOperationType, PNPushType, HttpMethod
from .models.consumer.common import PNStatus
from .errors import PNERR_JSON_NOT_SERIALIZABLE, PNERR_PERMISSION_MISSING
from .exceptions import PubNubException


def get_data_for_user(data):
    try:
        if 'message' in data and 'payload' in data:
            return {'message': data['message'], 'payload': data['payload']}
        else:
            return data
    except TypeError:
        return data


def write_value_as_string(data):
    try:
        if isinstance(data, str):
            return "\"%s\"" % data
        else:
            return json.dumps(data)
    except TypeError:
        raise PubNubException(
            pn_error=PNERR_JSON_NOT_SERIALIZABLE
        )


def url_encode(data):
    return urllib.parse.quote(data, safe="~").replace("+", "%2B")


def url_write(data):
    """ Just wraps url_encode(write_value_as_string()) """
    return url_encode(write_value_as_string(data))


def uuid():
    return str(u.uuid4())


def split_items(items_string):
    if len(items_string) == 0:
        return []
    else:
        return items_string.split(",")


def join_items(items_list):
    return ",".join(items_list)


def join_items_and_encode(items_list):
    return ",".join(url_encode(x) for x in items_list)


def join_channels(items_list):
    if len(items_list) == 0:
        return ","
    else:
        return join_items_and_encode(items_list)


def extend_list(existing_items, new_items):
    if isinstance(new_items, str):
        existing_items.extend(split_items(new_items))
    else:
        existing_items.extend(new_items)


def build_url(scheme, origin, path, params={}):
    return urllib.parse.urlunsplit((scheme, origin, path, params, ''))


def synchronized(func):
    func.__lock__ = threading.Lock()

    def synced_func(*args, **kws):
        with func.__lock__:
            return func(*args, **kws)

    return synced_func


def is_subscribed_event(status):
    assert isinstance(status, PNStatus)
    return status.category == PNStatusCategory.PNConnectedCategory


def is_unsubscribed_event(status):
    assert isinstance(status, PNStatus)
    return status.category == PNStatusCategory.PNAcknowledgmentCategory \
        and status.operation == PNOperationType.PNUnsubscribeOperation


def prepare_pam_arguments(unsorted_params):
    sorted_keys = sorted(unsorted_params)
    stringified_arguments = ""
    i = 0

    for key in sorted_keys:
        if i != 0:
            stringified_arguments += "&"

        stringified_arguments += (key + "=" + pam_encode(str(unsorted_params[key])))
        i += 1

    return stringified_arguments


def pam_encode(s_url):
    # !'()*~
    encoded = url_encode(s_url)
    if encoded is not None:
        encoded = (encoded.replace("*", "%2A")
                   .replace("!", "%21")
                   .replace("'", "%27")
                   .replace("(", "%28")
                   .replace(")", "%29")
                   .replace("[", "%5B")
                   .replace("]", "%5D")
                   .replace("~", "%7E"))

    return encoded


def sign_sha256(secret, sign_input):
    from base64 import urlsafe_b64encode

    sign = urlsafe_b64encode(hmac.new(
        secret.encode("utf-8"),
        sign_input.encode("utf-8"),
        sha256
    ).digest())

    return sign.decode("utf-8")


def push_type_to_string(push_type):
    if push_type == PNPushType.APNS:
        return "apns"
    elif push_type == PNPushType.GCM:
        return "gcm"
    elif push_type == PNPushType.MPNS:
        return "mpns"
    else:
        return ""


def strip_right(text, suffix):
    if not text.endswith(suffix):
        return text

    return text[:len(text) - len(suffix)]


def datetime_now():
    return datetime.datetime.now().strftime("%I:%M%p on %B %d, %Y")


def sign_request(endpoint, pn, custom_params, method, body):
    custom_params['timestamp'] = str(pn.timestamp())

    request_url = endpoint.build_path()

    encoded_query_string = prepare_pam_arguments(custom_params)

    is_v2_signature = not (request_url.startswith("/publish") and method == HttpMethod.POST)

    signed_input = ""
    if not is_v2_signature:
        signed_input += pn.config.subscribe_key + "\n"
        signed_input += pn.config.publish_key + "\n"
        signed_input += request_url + "\n"
        signed_input += encoded_query_string
    else:
        signed_input += HttpMethod.string(method).upper() + "\n"
        signed_input += pn.config.publish_key + "\n"
        signed_input += request_url + "\n"
        signed_input += encoded_query_string + "\n"
        if body is not None:
            signed_input += body

    signature = sign_sha256(pn.config.secret_key, signed_input)
    if is_v2_signature:
        signature = signature.rstrip("=")
        signature = "v2." + signature

    custom_params['signature'] = signature


def parse_resources(resource_list, resource_set_name, resources, patterns):
    if resource_list is not None:
        for pn_resource in resource_list:
            resource_object = {}

            if pn_resource.is_pattern_resource():
                determined_object = patterns
            else:
                determined_object = resources

            if resource_set_name in determined_object:
                determined_object[resource_set_name][pn_resource.get_id()] = calculate_bitmask(pn_resource)
            else:
                resource_object[pn_resource.get_id()] = calculate_bitmask(pn_resource)
                determined_object[resource_set_name] = resource_object

    if resource_set_name not in resources:
        resources[resource_set_name] = {}

    if resource_set_name not in patterns:
        patterns[resource_set_name] = {}


def calculate_bitmask(pn_resource):
    bit_sum = 0
    from .endpoints.access.grant_token import GrantToken

    if pn_resource.is_read() is True:
        bit_sum += GrantToken.READ

    if pn_resource.is_write() is True:
        bit_sum += GrantToken.WRITE

    if pn_resource.is_manage() is True:
        bit_sum += GrantToken.MANAGE

    if pn_resource.is_delete() is True:
        bit_sum += GrantToken.DELETE

    if pn_resource.is_create() is True:
        bit_sum += GrantToken.CREATE

    if bit_sum == 0:
        raise PubNubException(pn_error=PNERR_PERMISSION_MISSING)

    return bit_sum


def decode_utf8_dict(dic):
    if isinstance(dic, bytes):
        return dic.decode("utf-8")
    elif isinstance(dic, dict):
        new_dic = {}

        for key in dic:
            new_key = key
            if isinstance(key, bytes):
                new_key = key.decode("UTF-8")

            if new_key == "sig" and isinstance(dic[key], bytes):
                new_dic[new_key] = dic[key]
            else:
                new_dic[new_key] = decode_utf8_dict(dic[key])

        return new_dic
    elif isinstance(dic, list):
        new_l = []
        for e in dic:
            new_l.append(decode_utf8_dict(e))
        return new_l
    else:
        return dic



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.errors import PNERR_CHANNEL_OR_GROUP_MISSING
from pubnub.exceptions import PubNubException


class Subscribe(Endpoint):
    # /subscribe/<sub-key>/<channel>/<callback>/<timetoken>
    SUBSCRIBE_PATH = "/v2/subscribe/%s/%s/0"

    def __init__(self, pubnub):
        super(Subscribe, self).__init__(pubnub)
        self._channels = []
        self._groups = []

        self._region = None
        self._filter_expression = None
        self._timetoken = None
        self._with_presence = None

    def channels(self, channels):
        utils.extend_list(self._channels, channels)

        return self

    def channel_groups(self, groups):
        utils.extend_list(self._groups, groups)

        return self

    def timetoken(self, timetoken):
        self._timetoken = timetoken

        return self

    def filter_expression(self, expr):
        self._filter_expression = expr

        return self

    def region(self, region):
        self._region = region

        return self

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()

        if len(self._channels) == 0 and len(self._groups) == 0:
            raise PubNubException(pn_error=PNERR_CHANNEL_OR_GROUP_MISSING)

    def build_path(self):
        channels = utils.join_channels(self._channels)
        return Subscribe.SUBSCRIBE_PATH % (self.pubnub.config.subscribe_key, channels)

    def custom_params(self):
        params = {}

        if len(self._groups) > 0:
            params['channel-group'] = utils.join_items_and_encode(self._groups)

        if self._filter_expression is not None and len(self._filter_expression) > 0:
            params['filter-expr'] = utils.url_encode(self._filter_expression)

        if self._timetoken is not None:
            params['tt'] = str(self._timetoken)

        if self._region is not None:
            params['tr'] = self._region

        if not self.pubnub.config.heartbeat_default_values:
            params['heartbeat'] = self.pubnub.config.presence_timeout

        return params

    def create_response(self, envelope):
        return envelope

    def is_auth_required(self):
        return True

    def affected_channels(self):
        return self._channels

    def affected_channels_groups(self):
        return self._groups

    def request_timeout(self):
        return self.pubnub.config.subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNSubscribeOperation

    def name(self):
        return "Subscribe"



import logging
from abc import ABCMeta

from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_UUID_MISSING, PNERR_CHANNEL_MISSING
from pubnub.exceptions import PubNubException
from pubnub.models.consumer.objects_v2.page import Next, Previous

logger = logging.getLogger("pubnub")


class ObjectsEndpoint(Endpoint):
    __metaclass__ = ABCMeta

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)

    def is_auth_required(self):
        return True

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_specific_params()

    def validate_specific_params(self):
        pass

    def custom_params(self):
        params = {}
        inclusions = []

        if isinstance(self, IncludeCustomEndpoint):
            if self._include_custom:
                inclusions.append("custom")

        if isinstance(self, UUIDIncludeEndpoint):
            if self._uuid_details_level:
                if self._uuid_details_level == UUIDIncludeEndpoint.UUID:
                    inclusions.append("uuid")
                elif self._uuid_details_level == UUIDIncludeEndpoint.UUID_WITH_CUSTOM:
                    inclusions.append("uuid.custom")

        if isinstance(self, ChannelIncludeEndpoint):
            if self._channel_details_level:
                if self._channel_details_level == ChannelIncludeEndpoint.CHANNEL:
                    inclusions.append("channel")
                elif self._channel_details_level == ChannelIncludeEndpoint.CHANNEL_WITH_CUSTOM:
                    inclusions.append("channel.custom")

        if isinstance(self, ListEndpoint):
            if self._filter:
                params["filter"] = utils.url_encode(str(self._filter))

            if self._limit:
                params["limit"] = int(self._limit)

            if self._include_total_count:
                params["count"] = bool(self._include_total_count)

            if self._sort_keys:
                joined_sort_params_array = []
                for sort_key in self._sort_keys:
                    joined_sort_params_array.append("%s:%s" % (sort_key.key_str(), sort_key.dir_str()))

                params["sort"] = ",".join(joined_sort_params_array)

            if self._page:
                if isinstance(self._page, Next):
                    params["start"] = self._page.hash()
                elif isinstance(self._page, Previous):
                    params["end"] = self._page.hash()
                else:
                    raise ValueError()

        if len(inclusions) > 0:
            params["include"] = ",".join(inclusions)

        return params


class CustomAwareEndpoint:
    __metaclass__ = ABCMeta

    def __init__(self):
        self._custom = None

    def custom(self, custom):
        self._custom = dict(custom)
        return self


class ChannelEndpoint:
    __metaclass__ = ABCMeta

    def __init__(self):
        self._channel = None

    def channel(self, channel):
        self._channel = str(channel)
        return self

    def _validate_channel(self):
        if self._channel is None or len(self._channel) == 0:
            raise PubNubException(pn_error=PNERR_CHANNEL_MISSING)


class UuidEndpoint:
    __metaclass__ = ABCMeta

    def __init__(self):
        self._uuid = None

    def uuid(self, uuid):
        self._uuid = str(uuid)
        return self

    def _effective_uuid(self):
        if self._uuid is not None:
            return self._uuid
        else:
            return self.pubnub.config.uuid

    def _validate_uuid(self):
        if self._effective_uuid() is None or len(self._effective_uuid()) == 0:
            raise PubNubException(pn_error=PNERR_UUID_MISSING)


class ListEndpoint:
    __metaclass__ = ABCMeta

    def __init__(self):
        self._limit = None
        self._filter = None
        self._include_total_count = None
        self._sort_keys = None
        self._page = None

    def limit(self, limit):
        self._limit = int(limit)
        return self

    def filter(self, filter):
        self._filter = str(filter)
        return self

    def include_total_count(self, include_total_count):
        self._include_total_count = bool(include_total_count)
        return self

    def sort(self, *sort_keys):
        self._sort_keys = sort_keys
        return self

    def page(self, page):
        self._page = page
        return self


class IncludeCustomEndpoint:
    __metaclass__ = ABCMeta

    def __init__(self):
        self._include_custom = None

    def include_custom(self, include_custom):
        self._include_custom = bool(include_custom)
        return self


class UUIDIncludeEndpoint:
    __metaclass__ = ABCMeta

    UUID = 1
    UUID_WITH_CUSTOM = 2

    def __init__(self):
        self._uuid_details_level = None

    def include_uuid(self, uuid_details_level):
        self._uuid_details_level = uuid_details_level
        return self


class ChannelIncludeEndpoint:
    __metaclass__ = ABCMeta

    CHANNEL = 1
    CHANNEL_WITH_CUSTOM = 2

    def __init__(self):
        self._channel_details_level = None

    def include_channel(self, channel_details_level):
        self._channel_details_level = channel_details_level
        return self



from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, ChannelEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.objects_v2.channel import PNRemoveChannelMetadataResult


class RemoveChannel(ObjectsEndpoint, ChannelEndpoint):
    REMOVE_CHANNEL_PATH = "/v2/objects/%s/channels/%s"

    def __init__(self, pubnub):
        ObjectsEndpoint.__init__(self, pubnub)
        ChannelEndpoint.__init__(self)

    def build_path(self):
        return RemoveChannel.REMOVE_CHANNEL_PATH % (self.pubnub.config.subscribe_key, self._channel)

    def validate_specific_params(self):
        self._validate_channel()

    def create_response(self, envelope):
        return PNRemoveChannelMetadataResult(envelope)

    def operation_type(self):
        return PNOperationType.PNRemoveChannelMetadataOperation

    def name(self):
        return "Remove Channel"

    def http_method(self):
        return HttpMethod.DELETE



import logging
import time

from abc import ABCMeta, abstractmethod

from .endpoints.objects_v2.uuid.set_uuid import SetUuid
from .endpoints.objects_v2.channel.get_all_channels import GetAllChannels
from .endpoints.objects_v2.channel.get_channel import GetChannel
from .endpoints.objects_v2.channel.remove_channel import RemoveChannel
from .endpoints.objects_v2.channel.set_channel import SetChannel
from .endpoints.objects_v2.members.get_channel_members import GetChannelMembers
from .endpoints.objects_v2.members.manage_channel_members import ManageChannelMembers
from .endpoints.objects_v2.members.remove_channel_members import RemoveChannelMembers
from .endpoints.objects_v2.members.set_channel_members import SetChannelMembers
from .endpoints.objects_v2.memberships.get_memberships import GetMemberships
from .endpoints.objects_v2.memberships.manage_memberships import ManageMemberships
from .endpoints.objects_v2.memberships.remove_memberships import RemoveMemberships
from .endpoints.objects_v2.memberships.set_memberships import SetMemberships
from .endpoints.objects_v2.uuid.get_all_uuid import GetAllUuid
from .endpoints.objects_v2.uuid.get_uuid import GetUuid
from .endpoints.objects_v2.uuid.remove_uuid import RemoveUuid
from .managers import BasePathManager, TokenManager, TokenManagerProperties
from .builders import SubscribeBuilder
from .builders import UnsubscribeBuilder
from .endpoints.time import Time
from .endpoints.history import History
from .endpoints.access.audit import Audit
from .endpoints.access.grant import Grant
from .endpoints.access.grant_token import GrantToken
from .endpoints.access.revoke import Revoke
from .endpoints.channel_groups.add_channel_to_channel_group import AddChannelToChannelGroup
from .endpoints.channel_groups.list_channels_in_channel_group import ListChannelsInChannelGroup
from .endpoints.channel_groups.remove_channel_from_channel_group import RemoveChannelFromChannelGroup
from .endpoints.channel_groups.remove_channel_group import RemoveChannelGroup
from .endpoints.presence.get_state import GetState
from .endpoints.presence.heartbeat import Heartbeat
from .endpoints.presence.set_state import SetState
from .endpoints.pubsub.publish import Publish
from .endpoints.pubsub.fire import Fire
from .endpoints.presence.here_now import HereNow
from .endpoints.presence.where_now import WhereNow
from .endpoints.history_delete import HistoryDelete
from .endpoints.message_count import MessageCount
from .endpoints.signal import Signal
from .endpoints.fetch_messages import FetchMessages
from .endpoints.message_actions.add_message_action import AddMessageAction
from .endpoints.message_actions.get_message_actions import GetMessageActions
from .endpoints.message_actions.remove_message_action import RemoveMessageAction
from .endpoints.file_operations.list_files import ListFiles
from .endpoints.file_operations.delete_file import DeleteFile
from .endpoints.file_operations.get_file_url import GetFileDownloadUrl
from .endpoints.file_operations.fetch_upload_details import FetchFileUploadS3Data
from .endpoints.file_operations.send_file import SendFileNative
from .endpoints.file_operations.download_file import DownloadFileNative
from .endpoints.file_operations.publish_file_message import PublishFileMessage

from .endpoints.push.add_channels_to_push import AddChannelsToPush
from .endpoints.push.remove_channels_from_push import RemoveChannelsFromPush
from .endpoints.push.remove_device import RemoveDeviceFromPush
from .endpoints.push.list_push_provisions import ListPushProvisions
from .managers import TelemetryManager

logger = logging.getLogger("pubnub")


class PubNubCore:
    """A base class for PubNub Python API implementations"""
    SDK_VERSION = "5.1.1"
    SDK_NAME = "PubNub-Python"

    TIMESTAMP_DIVIDER = 1000
    MAX_SEQUENCE = 65535

    __metaclass__ = ABCMeta

    def __init__(self, config):
        self.config = config
        self.config.validate()
        self.headers = {
            'User-Agent': self.sdk_name
        }

        self._subscription_manager = None
        self._publish_sequence_manager = None
        self._telemetry_manager = TelemetryManager()
        self._base_path_manager = BasePathManager(config)
        self._token_manager = TokenManager()

    @property
    def base_origin(self):
        return self._base_path_manager.get_base_path()

    @property
    def sdk_name(self):
        return "%s%s/%s" % (PubNubCore.SDK_NAME, self.sdk_platform(), PubNubCore.SDK_VERSION)

    @abstractmethod
    def sdk_platform(self):
        pass

    @property
    def uuid(self):
        return self.config.uuid

    def add_listener(self, listener):
        self._validate_subscribe_manager_enabled()

        return self._subscription_manager.add_listener(listener)

    def remove_listener(self, listener):
        self._validate_subscribe_manager_enabled()

        return self._subscription_manager.remove_listener(listener)

    def get_subscribed_channels(self):
        self._validate_subscribe_manager_enabled()

        return self._subscription_manager.get_subscribed_channels()

    def get_subscribed_channel_groups(self):
        self._validate_subscribe_manager_enabled()

        return self._subscription_manager.get_subscribed_channel_groups()

    def add_channel_to_channel_group(self):
        return AddChannelToChannelGroup(self)

    def remove_channel_from_channel_group(self):
        return RemoveChannelFromChannelGroup(self)

    def list_channels_in_channel_group(self):
        return ListChannelsInChannelGroup(self)

    def remove_channel_group(self):
        return RemoveChannelGroup(self)

    def subscribe(self):
        return SubscribeBuilder(self._subscription_manager)

    def unsubscribe(self):
        return UnsubscribeBuilder(self._subscription_manager)

    def unsubscribe_all(self):
        return self._subscription_manager.unsubscribe_all()

    def reconnect(self):
        return self._subscription_manager.reconnect()

    def heartbeat(self):
        return Heartbeat(self)

    def set_state(self):
        return SetState(self, self._subscription_manager)

    def get_state(self):
        return GetState(self)

    def here_now(self):
        return HereNow(self)

    def where_now(self):
        return WhereNow(self)

    def publish(self):
        return Publish(self)

    def grant(self):
        return Grant(self)

    def grant_token(self):
        return GrantToken(self)

    def revoke(self):
        return Revoke(self)

    def audit(self):
        return Audit(self)

    # Push Related methods
    def list_push_channels(self):
        return ListPushProvisions(self)

    def add_channels_to_push(self):
        return AddChannelsToPush(self)

    def remove_channels_from_push(self):
        return RemoveChannelsFromPush(self)

    def remove_device_from_push(self):
        return RemoveDeviceFromPush(self)

    def history(self):
        return History(self)

    def message_counts(self):
        return MessageCount(self)

    def fire(self):
        return Fire(self)

    def signal(self):
        return Signal(self)

    def set_uuid_metadata(self):
        return SetUuid(self)

    def get_uuid_metadata(self):
        return GetUuid(self)

    def remove_uuid_metadata(self):
        return RemoveUuid(self)

    def get_all_uuid_metadata(self):
        return GetAllUuid(self)

    def set_channel_metadata(self):
        return SetChannel(self)

    def get_channel_metadata(self):
        return GetChannel(self)

    def remove_channel_metadata(self):
        return RemoveChannel(self)

    def get_all_channel_metadata(self):
        return GetAllChannels(self)

    def set_channel_members(self):
        return SetChannelMembers(self)

    def get_channel_members(self):
        return GetChannelMembers(self)

    def remove_channel_members(self):
        return RemoveChannelMembers(self)

    def manage_channel_members(self):
        return ManageChannelMembers(self)

    def set_memberships(self):
        return SetMemberships(self)

    def get_memberships(self):
        return GetMemberships(self)

    def remove_memberships(self):
        return RemoveMemberships(self)

    def manage_memberships(self):
        return ManageMemberships(self)

    def fetch_messages(self):
        return FetchMessages(self)

    def add_message_action(self):
        return AddMessageAction(self)

    def get_message_actions(self):
        return GetMessageActions(self)

    def remove_message_action(self):
        return RemoveMessageAction(self)

    def time(self):
        return Time(self)

    def delete_messages(self):
        return HistoryDelete(self)

    def set_token(self, token):
        self._token_manager.set_token(token)

    def set_tokens(self, tokens):
        self._token_manager.set_tokens(tokens)

    def get_token(self, tms_properties):
        return self._token_manager.get_token(tms_properties)

    def get_token_by_resource(self, resource_id, resource_type):
        return self._token_manager.get_token(TokenManagerProperties(
            resource_id=resource_id,
            resource_type=resource_type
        ))

    def get_tokens(self):
        return self._token_manager.get_tokens()

    def get_tokens_by_resource(self, resource_type):
        return self._token_manager.get_tokens_by_resource(resource_type)

    def send_file(self):
        if not self.sdk_platform():
            return SendFileNative(self)
        elif "Asyncio" in self.sdk_platform():
            from .endpoints.file_operations.send_file_asyncio import AsyncioSendFile
            return AsyncioSendFile(self)
        else:
            raise NotImplementedError

    def download_file(self):
        if not self.sdk_platform():
            return DownloadFileNative(self)
        elif "Asyncio" in self.sdk_platform():
            from .endpoints.file_operations.download_file_asyncio import DownloadFileAsyncio
            return DownloadFileAsyncio(self)
        else:
            raise NotImplementedError

    def list_files(self):
        return ListFiles(self)

    def get_file_url(self):
        return GetFileDownloadUrl(self)

    def delete_file(self):
        return DeleteFile(self)

    def _fetch_file_upload_s3_data(self):
        return FetchFileUploadS3Data(self)

    def publish_file_message(self):
        return PublishFileMessage(self)

    def decrypt(self, cipher_key, file):
        return self.config.file_crypto.decrypt(cipher_key, file)

    def encrypt(self, cipher_key, file):
        return self.config.file_crypto.encrypt(cipher_key, file)

    @staticmethod
    def timestamp():
        return int(time.time())

    def _validate_subscribe_manager_enabled(self):
        if self._subscription_manager is None:
            raise Exception("Subscription manager is not enabled for this instance")



class PNResource(object):

    def __init__(self, resource_name=None, resource_pattern=None):
        self._resource_name = resource_name
        self._resource_pattern = resource_pattern
        self._read = False
        self._write = False
        self._create = False
        self._manage = False
        self._delete = False

    def is_pattern_resource(self):
        return self._resource_pattern is not None

    def get_id(self):
        if self.is_pattern_resource():
            return self._resource_pattern

        return self._resource_name

    def is_read(self):
        return self._read

    def is_write(self):
        return self._write

    def is_create(self):
        return self._create

    def is_manage(self):
        return self._manage

    def is_delete(self):
        return self._delete



from pubnub.endpoints.file_operations.file_based_endpoint import FileOperationEndpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub import utils
from pubnub.models.consumer.file import PNPublishFileMessageResult
from pubnub.endpoints.mixins import TimeTokenOverrideMixin


class PublishFileMessage(FileOperationEndpoint, TimeTokenOverrideMixin):
    PUBLISH_FILE_MESSAGE = "/v1/files/publish-file/%s/%s/0/%s/0/%s"

    def __init__(self, pubnub):
        super(PublishFileMessage, self).__init__(pubnub)
        self._file_id = None
        self._file_name = None
        self._pubnub = pubnub
        self._message = None
        self._should_store = None
        self._ttl = 0
        self._meta = None
        self._cipher_key = None
        self._replicate = None
        self._ptto = None

    def meta(self, meta):
        self._meta = meta
        return self

    def should_store(self, should_store):
        self._should_store = bool(should_store)
        return self

    def cipher_key(self, cipher_key):
        self._cipher_key = cipher_key
        return self

    def message(self, message):
        self._message = message
        return self

    def file_id(self, file_id):
        self._file_id = file_id
        return self

    def ttl(self, ttl):
        self._ttl = ttl
        return self

    def file_name(self, file_name):
        self._file_name = file_name
        return self

    def _encrypt_message(self, message):
        if self._cipher_key or self._pubnub.config.cipher_key:
            return self._pubnub.config.crypto.encrypt(
                self._cipher_key or self._pubnub.config.cipher_key,
                utils.write_value_as_string(message)
            )
        else:
            return message

    def _build_message(self):
        message = {
            "message": self._message,
            "file": {
                "id": self._file_id,
                "name": self._file_name
            }
        }
        return self._encrypt_message(message)

    def build_path(self):
        message = self._build_message()
        return PublishFileMessage.PUBLISH_FILE_MESSAGE % (
            self.pubnub.config.publish_key,
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel),
            utils.url_write(message)
        )

    def http_method(self):
        return HttpMethod.GET

    def custom_params(self):
        params = TimeTokenOverrideMixin.custom_params(self)
        params.update({
            "meta": utils.url_write(self._meta),
            "ttl": self._ttl,
            "store": 1 if self._should_store else 0
        })
        return params

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()
        self.validate_file_name()
        self.validate_file_id()

    def create_response(self, envelope):
        return PNPublishFileMessageResult(envelope)

    def operation_type(self):
        return PNOperationType.PNSendFileAction

    def name(self):
        return "Sending file upload notification"



import logging
import json
import asyncio
import aiohttp
import math
import time
import urllib

from asyncio import Event, Queue, Semaphore
from yarl import URL

from pubnub.models.consumer.common import PNStatus
from .endpoints.presence.heartbeat import Heartbeat
from .endpoints.presence.leave import Leave
from .endpoints.pubsub.subscribe import Subscribe
from .pubnub_core import PubNubCore
from .workers import SubscribeMessageWorker
from .managers import SubscriptionManager, PublishSequenceManager, ReconnectionManager, TelemetryManager
from . import utils
from .structures import ResponseInfo, RequestOptions
from .enums import PNStatusCategory, PNHeartbeatNotificationOptions, PNOperationType, PNReconnectionPolicy
from .callbacks import SubscribeCallback, ReconnectionCallback
from .errors import (
    PNERR_SERVER_ERROR, PNERR_CLIENT_ERROR, PNERR_JSON_DECODING_FAILED,
    PNERR_REQUEST_CANCELLED, PNERR_CLIENT_TIMEOUT
)
from .exceptions import PubNubException

logger = logging.getLogger("pubnub")


class PubNubAsyncio(PubNubCore):
    """
    PubNub Python SDK for asyncio framework
    """

    def __init__(self, config, custom_event_loop=None):
        super(PubNubAsyncio, self).__init__(config)
        self.event_loop = custom_event_loop or asyncio.get_event_loop()

        self._connector = None
        self._session = None

        self._connector = aiohttp.TCPConnector(verify_ssl=True)
        self._session = aiohttp.ClientSession(
            loop=self.event_loop,
            conn_timeout=self.config.connect_timeout,
            connector=self._connector
        )

        if self.config.enable_subscribe:
            self._subscription_manager = AsyncioSubscriptionManager(self)

        self._publish_sequence_manager = AsyncioPublishSequenceManager(self.event_loop, PubNubCore.MAX_SEQUENCE)

        self._telemetry_manager = AsyncioTelemetryManager()

    async def set_connector(self, cn):
        await self._session.close()

        self._connector = cn

        self._session = aiohttp.ClientSession(
            loop=self.event_loop,
            conn_timeout=self.config.connect_timeout,
            connector=self._connector
        )

    async def stop(self):
        await self._session.close()
        if self._subscription_manager:
            self._subscription_manager.stop()

    def sdk_platform(self):
        return "-Asyncio"

    def request_sync(self, *args):
        raise NotImplementedError

    def request_deferred(self, *args):
        raise NotImplementedError

    async def request_result(self, options_func, cancellation_event):
        envelope = await self._request_helper(options_func, cancellation_event)
        return envelope.result

    async def request_future(self, options_func, cancellation_event):
        try:
            res = await self._request_helper(options_func, cancellation_event)
            return res
        except PubNubException as e:
            return PubNubAsyncioException(
                result=None,
                status=e.status
            )
        except asyncio.TimeoutError:
            return PubNubAsyncioException(
                result=None,
                status=options_func().create_status(
                    PNStatusCategory.PNTimeoutCategory,
                    None,
                    None,
                    exception=PubNubException(
                        pn_error=PNERR_CLIENT_TIMEOUT
                    )
                )
            )
        except asyncio.CancelledError:
            return PubNubAsyncioException(
                result=None,
                status=options_func().create_status(
                    PNStatusCategory.PNCancelledCategory,
                    None,
                    None,
                    exception=PubNubException(
                        pn_error=PNERR_REQUEST_CANCELLED
                    )
                )
            )
        except Exception as e:
            return PubNubAsyncioException(
                result=None,
                status=options_func().create_status(
                    PNStatusCategory.PNUnknownCategory,
                    None,
                    None,
                    e
                )
            )

    async def _request_helper(self, options_func, cancellation_event):
        """
        Query string should be provided as a manually serialized and encoded string.

        :param options_func:
        :param cancellation_event:
        :return:
        """
        if cancellation_event is not None:
            assert isinstance(cancellation_event, Event)

        options = options_func()
        assert isinstance(options, RequestOptions)

        create_response = options.create_response
        create_status = options.create_status
        create_exception = options.create_exception

        params_to_merge_in = {}

        if options.operation_type == PNOperationType.PNPublishOperation:
            params_to_merge_in['seqn'] = await self._publish_sequence_manager.get_next_sequence()

        options.merge_params_in(params_to_merge_in)

        if options.use_base_path:
            url = utils.build_url(self.config.scheme(), self.base_origin, options.path, options.query_string)
        else:
            url = utils.build_url(scheme="", origin="", path=options.path, params=options.query_string)

        url = URL(url, encoded=True)

        logger.debug("%s %s %s" % (options.method_string, url, options.data))

        if options.request_headers:
            self.headers.update(options.request_headers)

        try:
            start_timestamp = time.time()
            response = await asyncio.wait_for(
                self._session.request(
                    options.method_string,
                    url,
                    headers=self.headers,
                    data=options.data if options.data else None,
                    allow_redirects=options.allow_redirects
                ),
                options.request_timeout
            )
        except (asyncio.TimeoutError, asyncio.CancelledError):
            raise
        except Exception as e:
            logger.error("session.request exception: %s" % str(e))
            raise

        if not options.non_json_response:
            body = await response.text()
        else:
            if isinstance(response.content, bytes):
                body = response.content  # TODO: simplify this logic within the v5 release
            else:
                body = await response.read()

        if cancellation_event is not None and cancellation_event.is_set():
            return

        response_info = None
        status_category = PNStatusCategory.PNUnknownCategory

        if response:
            request_url = urllib.parse.urlparse(str(response.url))
            query = urllib.parse.parse_qs(request_url.query)
            uuid = None
            auth_key = None

            if 'uuid' in query and len(query['uuid']) > 0:
                uuid = query['uuid'][0]

            if 'auth_key' in query and len(query['auth_key']) > 0:
                auth_key = query['auth_key'][0]

            response_info = ResponseInfo(
                status_code=response.status,
                tls_enabled='https' == request_url.scheme,
                origin=request_url.hostname,
                uuid=uuid,
                auth_key=auth_key,
                client_request=None,
                client_response=response
            )

        # if body is not None and len(body) > 0 and not options.non_json_response:
        if body is not None and len(body) > 0:
            if options.non_json_response:
                data = body
            else:
                try:
                    data = json.loads(body)
                except ValueError:
                    if response.status == 599 and len(body) > 0:
                        data = body
                    else:
                        raise
                except TypeError:
                    try:
                        data = json.loads(body.decode("utf-8"))
                    except ValueError:
                        raise create_exception(
                            category=status_category,
                            response=response,
                            response_info=response_info,
                            exception=PubNubException(
                                pn_error=PNERR_JSON_DECODING_FAILED,
                                errormsg='json decode error',
                            )
                        )
        else:
            data = "N/A"

        logger.debug(data)

        if response.status not in (200, 307, 204):

            if response.status >= 500:
                err = PNERR_SERVER_ERROR
            else:
                err = PNERR_CLIENT_ERROR

            if response.status == 403:
                status_category = PNStatusCategory.PNAccessDeniedCategory

            if response.status == 400:
                status_category = PNStatusCategory.PNBadRequestCategory

            raise create_exception(
                category=status_category,
                response=data,
                response_info=response_info,
                exception=PubNubException(
                    errormsg=data,
                    pn_error=err,
                    status_code=response.status
                )
            )
        else:
            self._telemetry_manager.store_latency(time.time() - start_timestamp, options.operation_type)

            return AsyncioEnvelope(
                result=create_response(data) if not options.non_json_response else create_response(response, data),
                status=create_status(
                    PNStatusCategory.PNAcknowledgmentCategory,
                    data,
                    response_info,
                    None
                )
            )


class AsyncioReconnectionManager(ReconnectionManager):
    def __init__(self, pubnub):
        self._task = None
        super(AsyncioReconnectionManager, self).__init__(pubnub)

    async def _register_heartbeat_timer(self):
        while True:
            self._recalculate_interval()

            await asyncio.sleep(self._timer_interval)

            logger.debug("reconnect loop at: %s" % utils.datetime_now())

            try:
                await self._pubnub.time().future()
                self._connection_errors = 1
                self._callback.on_reconnect()
                break
            except Exception:
                if self._pubnub.config.reconnect_policy == PNReconnectionPolicy.EXPONENTIAL:
                    logger.debug("reconnect interval increment at: %s" % utils.datetime_now())
                    self._connection_errors += 1

    def start_polling(self):
        if self._pubnub.config.reconnect_policy == PNReconnectionPolicy.NONE:
            logger.warning("reconnection policy is disabled, please handle reconnection manually.")
            return

        self._task = asyncio.ensure_future(self._register_heartbeat_timer())

    def stop_polling(self):
        if self._task is not None and not self._task.cancelled():
            self._task.cancel()


class AsyncioPublishSequenceManager(PublishSequenceManager):
    def __init__(self, ioloop, provided_max_sequence):
        super(AsyncioPublishSequenceManager, self).__init__(provided_max_sequence)
        self._lock = asyncio.Lock()
        self._event_loop = ioloop

    async def get_next_sequence(self):
        async with self._lock:
            if self.max_sequence == self.next_sequence:
                self.next_sequence = 1
            else:
                self.next_sequence += 1

            return self.next_sequence


class AsyncioSubscriptionManager(SubscriptionManager):
    def __init__(self, pubnub_instance):
        subscription_manager = self

        self._message_worker = None
        self._message_queue = Queue()
        self._subscription_lock = Semaphore(1)
        self._subscribe_loop_task = None
        self._heartbeat_periodic_callback = None
        self._reconnection_manager = AsyncioReconnectionManager(pubnub_instance)

        super(AsyncioSubscriptionManager, self).__init__(pubnub_instance)
        self._start_worker()

        class AsyncioReconnectionCallback(ReconnectionCallback):
            def on_reconnect(self):
                subscription_manager.reconnect()

                pn_status = PNStatus()
                pn_status.category = PNStatusCategory.PNReconnectedCategory
                pn_status.error = False

                subscription_manager._subscription_status_announced = True
                subscription_manager._listener_manager.announce_status(pn_status)

        self._reconnection_listener = AsyncioReconnectionCallback()
        self._reconnection_manager.set_reconnection_listener(self._reconnection_listener)

    def _set_consumer_event(self):
        if not self._message_worker.cancelled():
            self._message_worker.cancel()

    def _message_queue_put(self, message):
        self._message_queue.put_nowait(message)

    def _start_worker(self):
        consumer = AsyncioSubscribeMessageWorker(
            self._pubnub,
            self._listener_manager,
            self._message_queue,
            None
        )
        self._message_worker = asyncio.ensure_future(
            consumer.run(),
            loop=self._pubnub.event_loop
        )

    def reconnect(self):
        # TODO: method is synchronized in Java
        self._should_stop = False
        self._subscribe_loop_task = asyncio.ensure_future(self._start_subscribe_loop())
        self._register_heartbeat_timer()

    def disconnect(self):
        # TODO: method is synchronized in Java
        self._should_stop = True
        self._stop_heartbeat_timer()
        self._stop_subscribe_loop()

    def stop(self):
        super(AsyncioSubscriptionManager, self).stop()
        self._reconnection_manager.stop_polling()
        if self._subscribe_loop_task and not self._subscribe_loop_task.cancelled():
            self._subscribe_loop_task.cancel()

    async def _start_subscribe_loop(self):
        self._stop_subscribe_loop()

        await self._subscription_lock.acquire()

        combined_channels = self._subscription_state.prepare_channel_list(True)
        combined_groups = self._subscription_state.prepare_channel_group_list(True)

        if len(combined_channels) == 0 and len(combined_groups) == 0:
            self._subscription_lock.release()
            return

        self._subscribe_request_task = asyncio.ensure_future(
            Subscribe(self._pubnub)
            .channels(combined_channels)
            .channel_groups(combined_groups)
            .timetoken(self._timetoken)
            .region(self._region)
            .filter_expression(self._pubnub.config.filter_expression)
            .future()
        )

        e = await self._subscribe_request_task

        if self._subscribe_request_task.cancelled():
            self._subscription_lock.release()
            return

        if e.is_error():
            if e.status and e.status.category == PNStatusCategory.PNCancelledCategory:
                self._subscription_lock.release()
                return

            if e.status and e.status.category == PNStatusCategory.PNTimeoutCategory:
                asyncio.ensure_future(self._start_subscribe_loop())
                self._subscription_lock.release()
                return

            logger.error("Exception in subscribe loop: %s" % str(e))

            if e.status and e.status.category == PNStatusCategory.PNAccessDeniedCategory:
                e.status.operation = PNOperationType.PNUnsubscribeOperation

            # TODO: raise error
            self._listener_manager.announce_status(e.status)

            self._reconnection_manager.start_polling()
            self._subscription_lock.release()
            self.disconnect()
            return
        else:
            self._handle_endpoint_call(e.result, e.status)
            self._subscription_lock.release()
            self._subscribe_loop_task = asyncio.ensure_future(self._start_subscribe_loop())

        self._subscription_lock.release()

    def _stop_subscribe_loop(self):
        if self._subscribe_request_task is not None and not self._subscribe_request_task.cancelled():
            self._subscribe_request_task.cancel()

    def _stop_heartbeat_timer(self):
        if self._heartbeat_periodic_callback is not None:
            self._heartbeat_periodic_callback.stop()

    def _register_heartbeat_timer(self):
        super(AsyncioSubscriptionManager, self)._register_heartbeat_timer()

        self._heartbeat_periodic_callback = AsyncioPeriodicCallback(
            self._perform_heartbeat_loop,
            self._pubnub.config.heartbeat_interval * 1000,
            self._pubnub.event_loop)
        if not self._should_stop:
            self._heartbeat_periodic_callback.start()

    async def _perform_heartbeat_loop(self):
        if self._heartbeat_call is not None:
            # TODO: cancel call
            pass

        cancellation_event = Event()
        state_payload = self._subscription_state.state_payload()
        presence_channels = self._subscription_state.prepare_channel_list(False)
        presence_groups = self._subscription_state.prepare_channel_group_list(False)

        if len(presence_channels) == 0 and len(presence_groups) == 0:
            return

        try:
            heartbeat_call = (Heartbeat(self._pubnub)
                              .channels(presence_channels)
                              .channel_groups(presence_groups)
                              .state(state_payload)
                              .cancellation_event(cancellation_event)
                              .future())

            envelope = await heartbeat_call

            heartbeat_verbosity = self._pubnub.config.heartbeat_notification_options
            if envelope.status.is_error:
                if heartbeat_verbosity in (PNHeartbeatNotificationOptions.ALL, PNHeartbeatNotificationOptions.FAILURES):
                    self._listener_manager.announce_status(envelope.status)
            else:
                if heartbeat_verbosity == PNHeartbeatNotificationOptions.ALL:
                    self._listener_manager.announce_status(envelope.status)

        except PubNubAsyncioException:
            pass
            # TODO: check correctness
            # if e.status is not None and e.status.category == PNStatusCategory.PNTimeoutCategory:
            #     self._start_subscribe_loop()
            # else:
            #     self._listener_manager.announce_status(e.status)
        finally:
            cancellation_event.set()

    def _send_leave(self, unsubscribe_operation):
        asyncio.ensure_future(self._send_leave_helper(unsubscribe_operation))

    async def _send_leave_helper(self, unsubscribe_operation):
        envelope = await Leave(self._pubnub) \
            .channels(unsubscribe_operation.channels) \
            .channel_groups(unsubscribe_operation.channel_groups).future()

        self._listener_manager.announce_status(envelope.status)


class AsyncioSubscribeMessageWorker(SubscribeMessageWorker):
    async def run(self):
        await self._take_message()

    async def _take_message(self):
        while True:
            try:
                msg = await self._queue.get()
                if msg is not None:
                    self._process_incoming_payload(msg)
                self._queue.task_done()
            except asyncio.CancelledError:
                logger.debug("Message Worker cancelled")
                break
            except Exception as e:
                logger.error("take message interrupted: %s" % str(e))
                raise


class AsyncioPeriodicCallback(object):
    def __init__(self, callback, callback_time, event_loop):
        self._callback = callback
        self._callback_time = callback_time
        self._event_loop = event_loop
        self._next_timeout = None
        self._running = False
        self._timeout = None

    def start(self):
        self._running = True
        self._next_timeout = self._event_loop.time()
        self._schedule_next()

    def stop(self):
        self._running = False
        if self._timeout is not None:
            self._timeout.cancel()
            self._timeout = None

    def _run(self):
        if not self._running:
            return
        try:
            asyncio.ensure_future(self._callback())
        except Exception:
            raise
        finally:
            self._schedule_next()

    def _schedule_next(self):
        current_time = self._event_loop.time()

        if self._next_timeout <= current_time:
            callback_time_sec = self._callback_time / 1000.0
            self._next_timeout += (math.floor(
                (current_time - self._next_timeout) / callback_time_sec) + 1) * callback_time_sec

        self._timeout = self._event_loop.call_at(self._next_timeout, self._run)


class AsyncioEnvelope(object):
    def __init__(self, result, status):
        self.result = result
        self.status = status

    @staticmethod
    def is_error():
        return False


class PubNubAsyncioException(Exception):
    def __init__(self, result, status):
        self.result = result
        self.status = status

    def __str__(self):
        return str(self.status.error_data.exception)

    @staticmethod
    def is_error():
        return True

    def value(self):
        return self.status.error_data.exception


class SubscribeListener(SubscribeCallback):
    def __init__(self):
        self.connected = False
        self.connected_event = Event()
        self.disconnected_event = Event()
        self.presence_queue = Queue()
        self.message_queue = Queue()
        self.error_queue = Queue()

    def status(self, pubnub, status):
        if utils.is_subscribed_event(status) and not self.connected_event.is_set():
            self.connected_event.set()
        elif utils.is_unsubscribed_event(status) and not self.disconnected_event.is_set():
            self.disconnected_event.set()
        elif status.is_error():
            self.error_queue.put_nowait(status.error_data.exception)

    def message(self, pubnub, message):
        self.message_queue.put_nowait(message)

    def presence(self, pubnub, presence):
        self.presence_queue.put_nowait(presence)

    async def _wait_for(self, coro):
        scc_task = asyncio.ensure_future(coro)
        err_task = asyncio.ensure_future(self.error_queue.get())

        await asyncio.wait([
            scc_task,
            err_task
        ], return_when=asyncio.FIRST_COMPLETED)

        if err_task.done() and not scc_task.done():
            if not scc_task.cancelled():
                scc_task.cancel()
            raise err_task.result()
        else:
            if not err_task.cancelled():
                err_task.cancel()
            return scc_task.result()

    async def wait_for_connect(self):
        if not self.connected_event.is_set():
            await self._wait_for(self.connected_event.wait())
        else:
            raise Exception("instance is already connected")

    async def wait_for_disconnect(self):
        if not self.disconnected_event.is_set():
            await self._wait_for(self.disconnected_event.wait())
        else:
            raise Exception("instance is already disconnected")

    async def wait_for_message_on(self, *channel_names):
        channel_names = list(channel_names)
        while True:
            try:
                env = await self._wait_for(self.message_queue.get())
                if env.channel in channel_names:
                    return env
                else:
                    continue
            finally:
                self.message_queue.task_done()

    async def wait_for_presence_on(self, *channel_names):
        channel_names = list(channel_names)
        while True:
            try:
                env = await self._wait_for(self.presence_queue.get())
                if env.channel in channel_names:
                    return env
                else:
                    continue
            finally:
                self.presence_queue.task_done()


class AsyncioTelemetryManager(TelemetryManager):
    def __init__(self):
        TelemetryManager.__init__(self)
        self._timer = AsyncioPeriodicCallback(
            self._start_clean_up_timer,
            self.CLEAN_UP_INTERVAL * self.CLEAN_UP_INTERVAL_MULTIPLIER,
            asyncio.get_event_loop())
        self._timer.start()

    async def _start_clean_up_timer(self):
        self.clean_up_telemetry_data()

    def _stop_clean_up_timer(self):
        self._timer.stop()



from pubnub.models.consumer.v3.pn_resource import PNResource


class Channel(PNResource):

    def __init__(self, resource_name=None, resource_pattern=None):
        super(Channel, self).__init__(resource_name, resource_pattern)

    @staticmethod
    def id(channel_id):
        channel = Channel(resource_name=channel_id)
        return channel

    @staticmethod
    def pattern(channel_pattern):
        channel = Channel(resource_pattern=channel_pattern)
        return channel

    def read(self):
        self._read = True
        return self

    def write(self):
        self._write = True
        return self

    def delete(self):
        self._delete = True
        return self



import logging
import threading
import requests
import json  # noqa # pylint: disable=W0611
import urllib

from requests import Session
from requests.adapters import HTTPAdapter

from pubnub import utils
from pubnub.enums import PNStatusCategory
from pubnub.errors import PNERR_CLIENT_ERROR, PNERR_UNKNOWN_ERROR, PNERR_TOO_MANY_REDIRECTS_ERROR,\
    PNERR_CLIENT_TIMEOUT, PNERR_HTTP_ERROR, PNERR_CONNECTION_ERROR
from pubnub.errors import PNERR_SERVER_ERROR
from pubnub.exceptions import PubNubException
from pubnub.request_handlers.base import BaseRequestHandler
from pubnub.structures import RequestOptions, PlatformOptions, ResponseInfo, Envelope

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


logger = logging.getLogger("pubnub")


class RequestsRequestHandler(BaseRequestHandler):
    """ PubNub Python SDK Native requests handler based on `requests` HTTP library. """
    ENDPOINT_THREAD_COUNTER = 0

    def __init__(self, pubnub):
        self.session = Session()

        self.session.mount('http://%s' % pubnub.config.origin, HTTPAdapter(max_retries=1, pool_maxsize=500))
        self.session.mount('https://%s' % pubnub.config.origin, HTTPAdapter(max_retries=1, pool_maxsize=500))
        self.session.mount('http://%s/v2/subscribe' % pubnub.config.origin, HTTPAdapter(pool_maxsize=500))
        self.session.mount('https://%s/v2/subscribe' % pubnub.config.origin, HTTPAdapter(pool_maxsize=500))

        self.pubnub = pubnub

    def sync_request(self, platform_options, endpoint_call_options):
        return self._build_envelope(platform_options, endpoint_call_options)

    def async_request(self, endpoint_name, platform_options, endpoint_call_options, callback, cancellation_event):
        call = Call()

        if cancellation_event is None:
            cancellation_event = threading.Event()

        def callback_to_invoke_in_separate_thread():
            try:
                envelope = self._build_envelope(platform_options, endpoint_call_options)
                if cancellation_event is not None and cancellation_event.isSet():
                    # Since there are no way to affect on ongoing request it's response will
                    # be just ignored on cancel call
                    return

                callback(envelope)
            except PubNubException as e:
                logger.error("Async request PubNubException. %s" % str(e))
                callback(Envelope(
                    result=None,
                    status=endpoint_call_options.create_status(
                        category=PNStatusCategory.PNBadRequestCategory,
                        response=None,
                        response_info=None,
                        exception=e)))
            except Exception as e:
                logger.error("Async request Exception. %s" % str(e))
                callback(Envelope(
                    result=None,
                    status=endpoint_call_options.create_status(
                        category=PNStatusCategory.PNInternalExceptionCategory,
                        response=None,
                        response_info=None,
                        exception=e)))
            finally:
                call.executed_cb()

        self.execute_callback_in_separate_thread(
            callback_to_invoke_in_separate_thread,
            endpoint_name,
            call,
            cancellation_event
        )

    def execute_callback_in_separate_thread(
            self, callback_to_invoke_in_another_thread, operation_name, call_obj, cancellation_event
    ):
        client = AsyncHTTPClient(callback_to_invoke_in_another_thread)

        thread = threading.Thread(
            target=client.run,
            name="Thread-%s-%d" % (operation_name, ++RequestsRequestHandler.ENDPOINT_THREAD_COUNTER)
        )
        thread.setDaemon(self.pubnub.config.daemon)
        thread.start()

        call_obj.thread = thread
        call_obj.cancellation_event = cancellation_event

        return call_obj

    def async_file_based_operation(self, func, callback, operation_name, cancellation_event=None):
        call = Call()

        if cancellation_event is None:
            cancellation_event = threading.Event()

        def callback_to_invoke_in_separate_thread():
            try:
                envelope = func()
                callback(envelope.result, envelope.status)
            except Exception as e:
                logger.error("Async file upload request Exception. %s" % str(e))
                callback(
                    Envelope(result=None, status=e)
                )
            finally:
                call.executed_cb()

        self.execute_callback_in_separate_thread(
            callback_to_invoke_in_separate_thread,
            operation_name,
            call,
            cancellation_event
        )

        return call

    def _build_envelope(self, p_options, e_options):
        """ A wrapper for _invoke_url to separate request logic """

        status_category = PNStatusCategory.PNUnknownCategory
        response_info = None

        url_base_path = self.pubnub.base_origin if e_options.use_base_path else None
        try:
            res = self._invoke_request(p_options, e_options, url_base_path)
        except PubNubException as e:
            if e._pn_error is PNERR_CONNECTION_ERROR:
                status_category = PNStatusCategory.PNUnexpectedDisconnectCategory
            elif e._pn_error is PNERR_CLIENT_TIMEOUT:
                status_category = PNStatusCategory.PNTimeoutCategory

            return Envelope(
                result=None,
                status=e_options.create_status(
                    category=status_category,
                    response=None,
                    response_info=response_info,
                    exception=e))

        if res is not None:
            url = urllib.parse.urlparse(res.url)
            query = urllib.parse.parse_qs(url.query)
            uuid = None
            auth_key = None

            if 'uuid' in query and len(query['uuid']) > 0:
                uuid = query['uuid'][0]

            if 'auth_key' in query and len(query['auth_key']) > 0:
                auth_key = query['auth_key'][0]

            response_info = ResponseInfo(
                status_code=res.status_code,
                tls_enabled='https' == url.scheme,
                origin=url.hostname,
                uuid=uuid,
                auth_key=auth_key,
                client_request=res.request
            )

        if not res.ok:
            if res.status_code == 403:
                status_category = PNStatusCategory.PNAccessDeniedCategory

            if res.status_code == 400:
                status_category = PNStatusCategory.PNBadRequestCategory

            if res.text is None:
                text = "N/A"
            else:
                text = res.text

            if res.status_code >= 500:
                err = PNERR_SERVER_ERROR
            else:
                err = PNERR_CLIENT_ERROR
            try:
                response = res.json()
            except JSONDecodeError:
                response = None
            return Envelope(
                result=None,
                status=e_options.create_status(
                    category=status_category,
                    response=response,
                    response_info=response_info,
                    exception=PubNubException(
                        pn_error=err,
                        errormsg=text,
                        status_code=res.status_code
                    )))
        else:
            if e_options.non_json_response:
                response = res
            else:
                response = res.json()

            return Envelope(
                result=e_options.create_response(response),
                status=e_options.create_status(
                    category=PNStatusCategory.PNAcknowledgmentCategory,
                    response=response,
                    response_info=response_info,
                    exception=None
                )
            )

    def _invoke_request(self, p_options, e_options, base_origin):
        assert isinstance(p_options, PlatformOptions)
        assert isinstance(e_options, RequestOptions)

        if base_origin:
            url = p_options.pn_config.scheme() + "://" + base_origin + e_options.path
        else:
            url = e_options.path

        if e_options.request_headers:
            p_options.update(e_options.request_headers)

        args = {
            "method": e_options.method_string,
            "headers": p_options.headers,
            "url": url,
            "params": e_options.query_string,
            "timeout": (e_options.connect_timeout, e_options.request_timeout),
            "allow_redirects": e_options.allow_redirects
        }

        if e_options.is_post() or e_options.is_patch():
            args["data"] = e_options.data
            args["files"] = e_options.files
            logger.debug("%s %s %s" % (
                e_options.method_string,
                utils.build_url(
                    p_options.pn_config.scheme(),
                    base_origin,
                    e_options.path,
                    e_options.query_string), e_options.data))
        else:
            logger.debug("%s %s" % (
                e_options.method_string,
                utils.build_url(
                    p_options.pn_config.scheme(),
                    base_origin,
                    e_options.path,
                    e_options.query_string)))

        try:
            res = self.session.request(**args)
            logger.debug("GOT %s" % res.text)
        except requests.exceptions.ConnectionError as e:
            raise PubNubException(
                pn_error=PNERR_CONNECTION_ERROR,
                errormsg=str(e)
            )
        except requests.exceptions.HTTPError as e:
            raise PubNubException(
                pn_error=PNERR_HTTP_ERROR,
                errormsg=str(e)
            )
        except requests.exceptions.Timeout as e:
            raise PubNubException(
                pn_error=PNERR_CLIENT_TIMEOUT,
                errormsg=str(e)
            )
        except requests.exceptions.TooManyRedirects as e:
            raise PubNubException(
                pn_error=PNERR_TOO_MANY_REDIRECTS_ERROR,
                errormsg=str(e)
            )
        except Exception as e:
            raise PubNubException(
                pn_error=PNERR_UNKNOWN_ERROR,
                errormsg=str(e)
            )

        return res


class AsyncHTTPClient:
    """A wrapper for threaded calls"""

    def __init__(self, callback_to_invoke):
        self._callback_to_invoke = callback_to_invoke

    def run(self):
        self._callback_to_invoke()


class Call(object):
    """
    A platform dependent representation of async PubNub method call
    """

    def __init__(self):
        self.thread = None
        self.cancellation_event = None
        self.is_executed = False
        self.is_canceled = False

    def cancel(self):
        """
        Set Event flag to stop thread on timeout. This will not stop thread immediately, it will stopped
          only after ongoing request will be finished
        :return: nothing
        """
        if self.cancellation_event is not None:
            self.cancellation_event.set()
        self.is_canceled = True

    def join(self):
        if isinstance(self.thread, threading.Thread):
            self.thread.join()

    def executed_cb(self):
        self.is_executed = True



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_GROUP_MISSING
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.channel_group import PNChannelGroupsRemoveGroupResult


class RemoveChannelGroup(Endpoint):
    # /v1/channel-registration/sub-key/<sub_key>/channel-group/<group_name>/remove
    REMOVE_PATH = "/v1/channel-registration/sub-key/%s/channel-group/%s/remove"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channel_group = None

    def channel_group(self, channel_group):
        self._channel_group = channel_group

        return self

    def custom_params(self):
        return {}

    def build_path(self):
        return RemoveChannelGroup.REMOVE_PATH % (
            self.pubnub.config.subscribe_key, utils.url_encode(self._channel_group))

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()

        if not isinstance(self._channel_group, str) or len(self._channel_group) == 0:
            raise PubNubException(pn_error=PNERR_GROUP_MISSING)

    def is_auth_required(self):
        return True

    def create_response(self, envelope):
        return PNChannelGroupsRemoveGroupResult()

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNRemoveGroupOperation

    def name(self):
        return "RemoveChannelGroup"



from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, \
    IncludeCustomEndpoint, UuidEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.objects_v2.uuid import PNGetUUIDMetadataResult


class GetUuid(ObjectsEndpoint, UuidEndpoint, IncludeCustomEndpoint):
    GET_UID_PATH = "/v2/objects/%s/uuids/%s"

    def __init__(self, pubnub):
        ObjectsEndpoint.__init__(self, pubnub)
        UuidEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)

    def build_path(self):
        return GetUuid.GET_UID_PATH % (self.pubnub.config.subscribe_key, self._effective_uuid())

    def validate_specific_params(self):
        self._validate_uuid()

    def create_response(self, envelope):
        return PNGetUUIDMetadataResult(envelope)

    def operation_type(self):
        return PNOperationType.PNGetUuidMetadataOperation

    def name(self):
        return "Get UUID"

    def http_method(self):
        return HttpMethod.GET



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_MESSAGE_ACTION_VALUE_MISSING, PNERR_MESSAGE_ACTION_TYPE_MISSING, \
    PNERR_MESSAGE_TIMETOKEN_MISSING, PNERR_MESSAGE_ACTION_MISSING
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.message_actions import PNAddMessageActionResult


class AddMessageAction(Endpoint):
    ADD_MESSAGE_ACTION_PATH = "/v1/message-actions/%s/channel/%s/message/%s"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channel = None
        self._message_action = None

    def channel(self, channel):
        self._channel = str(channel)
        return self

    def message_action(self, message_action):
        self._message_action = message_action
        return self

    def custom_params(self):
        return {}

    def build_data(self):
        params = {
            'type': self._message_action.type,
            'value': self._message_action.value
        }

        return utils.write_value_as_string(params)

    def build_path(self):
        return AddMessageAction.ADD_MESSAGE_ACTION_PATH % (
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel),
            self._message_action.message_timetoken
        )

    def http_method(self):
        return HttpMethod.POST

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()
        self.validate_message_action()

    def create_response(self, envelope):  # pylint: disable=W0221
        return PNAddMessageActionResult(envelope['data'])

    def is_auth_required(self):
        return True

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNAddMessageAction

    def name(self):
        return "Add message action"

    def validate_message_action(self):
        if self._message_action is None:
            raise PubNubException(pn_error=PNERR_MESSAGE_ACTION_MISSING)

        if self._message_action.message_timetoken is None:
            raise PubNubException(pn_error=PNERR_MESSAGE_TIMETOKEN_MISSING)

        if self._message_action.type is None or len(self._message_action.type) == 0:
            raise PubNubException(pn_error=PNERR_MESSAGE_ACTION_TYPE_MISSING)

        if self._message_action.value is None or len(self._message_action.value) == 0:
            raise PubNubException(pn_error=PNERR_MESSAGE_ACTION_VALUE_MISSING)



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.history import PNHistoryResult


class History(Endpoint):
    HISTORY_PATH = "/v2/history/sub-key/%s/channel/%s"
    MAX_COUNT = 100

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channel = None
        self._start = None
        self._end = None
        self._reverse = None
        self._count = None
        self._include_timetoken = None
        self._include_meta = None

    def channel(self, channel):
        self._channel = channel
        return self

    def start(self, start):
        assert isinstance(start, int)
        self._start = start
        return self

    def end(self, end):
        assert isinstance(end, int)
        self._end = end
        return self

    def reverse(self, reverse):
        assert isinstance(reverse, bool)
        self._reverse = reverse
        return self

    def count(self, count):
        assert isinstance(count, int)
        self._count = count
        return self

    def include_timetoken(self, include_timetoken):
        assert isinstance(include_timetoken, bool)
        self._include_timetoken = include_timetoken
        return self

    def include_meta(self, include_meta):
        assert isinstance(include_meta, bool)
        self._include_meta = include_meta
        return self

    def custom_params(self):
        params = {}

        if self._start is not None:
            params['start'] = str(self._start)

        if self._end is not None:
            params['end'] = str(self._end)

        if self._count is not None and 0 < self._count <= History.MAX_COUNT:
            params['count'] = str(self._count)
        else:
            params['count'] = '100'

        if self._reverse is not None:
            params['reverse'] = "true" if self._reverse else "false"

        if self._include_timetoken is not None:
            params['include_token'] = "true" if self._include_timetoken else "false"

        if self._include_meta is not None:
            params['include_meta'] = "true" if self._include_meta else "false"

        return params

    def build_path(self):
        return History.HISTORY_PATH % (
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel)
        )

    def http_method(self):
        return HttpMethod.GET

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()

    def create_response(self, envelope):
        return PNHistoryResult.from_json(
            json_input=envelope,
            crypto=self.pubnub.config.crypto,
            include_timetoken=self._include_timetoken,
            include_meta=self._include_meta,
            cipher=self.pubnub.config.cipher_key)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNHistoryOperation

    def name(self):
        return "History"



from setuptools import setup, find_packages

setup(
    name='pubnub',
    version='5.1.1',
    description='PubNub Real-time push service in the cloud',
    author='PubNub',
    author_email='support@pubnub.com',
    url='http://pubnub.com',
    packages=find_packages(exclude=("examples*", 'tests*')),
    license='MIT',
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ),
    install_requires=[
        'pycryptodomex>=3.3',
        'requests>=2.4',
        'cbor2',
        'aiohttp'
    ],
    zip_safe=False,
)



class PNGetFilesResult:
    def __init__(self, result):
        self.data = result['data']
        self.count = result.get('count', None)
        self.next = result.get('next', None)
        self.prev = result.get('prev', None)

    def __str__(self):
        return "Get files success with data: %s" % self.data


class PNDeleteFileResult:
    def __init__(self, result):
        self.status = result['status']

    def __str__(self):
        return "Delete files success with status: %s" % self.status


class PNGetFileDownloadURLResult:
    def __init__(self, result, data=None):
        self.file_url = result.headers["Location"]

    def __str__(self):
        return "Get file URL success with status: %s" % self.status


class PNFetchFileUploadS3DataResult:
    def __init__(self, result):
        self.name = result["data"]["name"]
        self.file_id = result["data"]["id"]
        self.data = result["file_upload_request"]

    def __str__(self):
        return "Fetch file upload S3 data success with status: %s" % self.status


class PNDownloadFileResult:
    def __init__(self, result):
        self.data = result

    def __str__(self):
        return "Downloading file success with status: %s" % self.status


class PNSendFileResult:
    def __init__(self, result, file_upload_data):
        self.name = file_upload_data.result.name
        self.file_id = file_upload_data.result.file_id

    def __str__(self):
        return "Sending file success with status: %s" % self.status


class PNPublishFileMessageResult:
    def __init__(self, result):
        self.timestamp = result[2]

    def __str__(self):
        return "Sending file notification success with status: %s" % self.status



class PNHereNowResult:
    def __init__(self, total_channels, total_occupancy, channels):
        assert isinstance(total_channels, int)
        assert isinstance(total_occupancy, int)

        self.total_channels = total_channels
        self.total_occupancy = total_occupancy
        self.channels = channels

    def __str__(self):
        return "HereNow Result total occupancy: %d, total channels: %d" % (self.total_occupancy, self.total_channels)

    @classmethod
    def from_json(cls, envelope, channel_names):
        # multiple
        if 'payload' in envelope and isinstance(envelope['payload'], dict):
            json_input = envelope['payload']

            channels = []
            if len(json_input['channels']) > 0:
                for channel_name, raw_data in json_input['channels'].items():
                    channels.append(PNHereNowChannelData.from_json(channel_name, raw_data))
                return PNHereNowResult(
                    total_channels=int(json_input['total_channels']),
                    total_occupancy=int(json_input['total_occupancy']),
                    channels=channels)
            elif len(channel_names) == 1:
                return PNHereNowResult(
                    total_channels=int(1),
                    total_occupancy=int(json_input['total_occupancy']),
                    channels=[PNHereNowChannelData(channel_names[0], 0, [])]
                )
            else:
                return PNHereNowResult(
                    total_channels=int(json_input['total_channels']),
                    total_occupancy=int(json_input['total_occupancy']),
                    channels={}
                )
        # empty
        elif 'occupancy' in envelope and int(envelope['occupancy']) == 0:
            return PNHereNowResult(
                total_channels=int(1),
                total_occupancy=int(envelope['occupancy']),
                channels=[PNHereNowChannelData(channel_names[0], 0, [])]
            )
        # single
        elif 'uuids' in envelope and isinstance(envelope['uuids'], list):
            occupants = []
            for user in envelope['uuids']:
                if isinstance(user, str):
                    occupants.append(PNHereNowOccupantsData(user, None))
                else:
                    state = user['state'] if 'state' in user else None
                    occupants.append(PNHereNowOccupantsData(user['uuid'], state))

            return PNHereNowResult(
                total_channels=1,
                total_occupancy=int(envelope['occupancy']),
                channels=[
                    PNHereNowChannelData(
                        channel_name=channel_names[0],
                        occupancy=envelope['occupancy'],
                        occupants=occupants
                    )
                ])
        else:
            return PNHereNowResult(
                total_channels=1,
                total_occupancy=int(envelope['occupancy']),
                channels=[
                    PNHereNowChannelData(
                        channel_name=channel_names[0],
                        occupancy=envelope['occupancy'],
                        occupants=[]
                    )
                ])


class PNHereNowChannelData(object):
    def __init__(self, channel_name, occupancy, occupants):
        self.channel_name = channel_name
        self.occupancy = occupancy
        self.occupants = occupants

    def __str__(self):
        return "HereNow Channel Data for channel '%s': occupancy: %d, occupants: %d" \
               % (self.channel_name, self.occupancy, self.occupants)

    @classmethod
    def from_json(cls, name, json_input):
        if 'uuids' in json_input:
            occupants = []
            for user in json_input['uuids']:
                if isinstance(user, dict) and len(user) > 0:
                    if 'state' in user:
                        occupants.append(PNHereNowOccupantsData(user['uuid'], user['state']))
                    else:
                        occupants.append(PNHereNowOccupantsData(user['uuid'], None))
                else:
                    occupants.append(PNHereNowOccupantsData(user, None))
        else:
            occupants = None

        return PNHereNowChannelData(
            channel_name=name,
            occupancy=int(json_input['occupancy']),
            occupants=occupants
        )


class PNHereNowOccupantsData(object):
    def __init__(self, uuid, state):
        self.uuid = uuid
        self.state = state

    def __str__(self):
        return "HereNow Occupants Data for '%s': %s" % (self.uuid, self.state)


class PNWhereNowResult(object):
    def __init__(self, channels):
        assert isinstance(channels, (list, tuple))
        self.channels = channels

    def __str__(self):
        return "User is currently subscribed to %s" % ", ".join(self.channels)

    @classmethod
    def from_json(cls, json_input):
        return PNWhereNowResult(json_input['payload']['channels'])


class PNSetStateResult(object):
    def __init__(self, state):
        assert isinstance(state, dict)
        self.state = state

    def __str__(self):
        return "New state %s successfully set" % self.state


class PNGetStateResult(object):
    def __init__(self, channels):
        assert isinstance(channels, dict)
        self.channels = channels

    def __str__(self):
        return "Current state is %s" % self.channels



from .enums import HttpMethod


class RequestOptions(object):
    def __init__(
            self, path, params_callback,
            method, request_timeout, connect_timeout,
            create_response, create_status, create_exception,
            operation_type, data=None, sort_arguments=False,
            allow_redirects=True, use_base_path=None, files=None,
            request_headers=None, non_json_response=False
    ):
        assert len(path) > 0
        assert callable(params_callback)
        assert isinstance(method, int)
        assert isinstance(request_timeout, int)
        assert isinstance(connect_timeout, int)
        if not (method is HttpMethod.GET or method is HttpMethod.POST or method is HttpMethod.DELETE
                or method is HttpMethod.PATCH):  # noqa
            raise AssertionError()

        self.params = None
        self.path = path
        self.params_callback = params_callback
        self._method = method
        self.request_timeout = request_timeout
        self.connect_timeout = connect_timeout
        # TODO: rename 'data' => 'body'
        self.data = data
        self.files = files
        self.body = data
        self.sort_params = sort_arguments

        self.create_response = create_response
        self.create_status = create_status
        self.create_exception = create_exception
        self.operation_type = operation_type
        self.allow_redirects = allow_redirects
        self.use_base_path = use_base_path
        self.request_headers = request_headers
        self.non_json_response = non_json_response

    def merge_params_in(self, params_to_merge_in):
        self.params = self.params_callback(params_to_merge_in)

    @property
    def method_string(self):
        return HttpMethod.string(self._method)

    def is_post(self):
        return self._method is HttpMethod.POST

    def is_patch(self):
        return self._method is HttpMethod.PATCH

    def query_list(self):
        """ All query keys and values should be already encoded inside a build_params() method"""
        s = []

        for k, v in self.params.items():
            s.append(str(k) + "=" + str(v))

        if self.sort_params:
            return sorted(s)
        else:
            return s

    @property
    def query_string(self):
        return str('&'.join(self.query_list()))

    def __str__(self):
        return "path: {0}, qs: {1}".format(self.path, self.query_string)


class PlatformOptions(object):
    def __init__(self, headers, pn_config):
        self.headers = headers
        self.pn_config = pn_config


class ResponseInfo(object):
    def __init__(self, status_code, tls_enabled, origin, uuid, auth_key, client_request, client_response=None):
        self.status_code = status_code
        self.tls_enabled = tls_enabled
        self.origin = origin
        self.uuid = uuid
        self.auth_key = auth_key
        self.client_request = client_request
        self.client_response = client_response


class Envelope(object):
    def __init__(self, result, status):
        self.result = result
        self.status = status
